@@ -54,7 +54,7 @@
 #include "ogrsf_frmts/ogrsf_frmts.h"
 #include "ogrsf_frmts/vrt/ogr_vrt.h"
 
-CPL_CVSID("$Id: ogrvrtlayer.cpp 3798cbe48457b7127606931896549f26507469db 2021-04-09 15:04:16 +0200 Even Rouault $")
+CPL_CVSID("$Id$")
 
 #define UNSUPPORTED_OP_READ_ONLY \
     "%s : unsupported operation on a read-only datasource."
@@ -248,7 +248,7 @@ bool OGRVRTLayer::FastInitialize( CPLXMLNode *psLTreeIn,
             OGRSpatialReference oSRS;
             oSRS.SetAxisMappingStrategy(OAMS_TRADITIONAL_GIS_ORDER);
 
-            if( oSRS.SetFromUserInput(pszLayerSRS) != OGRERR_NONE )
+            if( oSRS.SetFromUserInput(pszLayerSRS, OGRSpatialReference::SET_FROM_USER_INPUT_LIMITATIONS_get()) != OGRERR_NONE )
             {
                 CPLError(CE_Failure, CPLE_AppDefined,
                          "Failed to import LayerSRS `%s'.", pszLayerSRS);
@@ -474,7 +474,7 @@ bool OGRVRTLayer::ParseGeometryField(CPLXMLNode *psNode,
             OGRSpatialReference oSRS;
             oSRS.SetAxisMappingStrategy(OAMS_TRADITIONAL_GIS_ORDER);
 
-            if( oSRS.SetFromUserInput(pszSRS) != OGRERR_NONE )
+            if( oSRS.SetFromUserInput(pszSRS, OGRSpatialReference::SET_FROM_USER_INPUT_LIMITATIONS_get()) != OGRERR_NONE )
             {
                 CPLError(CE_Failure, CPLE_AppDefined,
                          "Failed to import SRS `%s'.", pszSRS);
@@ -551,10 +551,9 @@ bool OGRVRTLayer::FullInitialize()
 
     // Figure out the data source name.  It may be treated relative
     // to vrt filename, but normally it is used directly.
-    char *pszSrcDSName =
-        const_cast<char *>(CPLGetXMLValue(psLTree, "SrcDataSource", nullptr));
+    std::string osSrcDSName = CPLGetXMLValue(psLTree, "SrcDataSource", "");
 
-    if( pszSrcDSName == nullptr )
+    if( osSrcDSName.empty() )
     {
         CPLError(CE_Failure, CPLE_AppDefined,
                  "Missing SrcDataSource for layer %s.", osName.c_str());
@@ -570,31 +569,26 @@ bool OGRVRTLayer::FullInitialize()
              i++ )
         {
             const char *pszPrefix = apszPrefixes[i];
-            if( EQUALN(pszSrcDSName, pszPrefix, strlen(pszPrefix)) )
+            if( EQUALN(osSrcDSName.c_str(), pszPrefix, strlen(pszPrefix)) )
             {
-                const char *pszLastPart = strrchr(pszSrcDSName, ':') + 1;
+                auto nLastPart = osSrcDSName.find(':') + 1;
                 // CSV:z:/foo.xyz
-                if( (pszLastPart[0] == '/' || pszLastPart[0] == '\\') &&
-                    pszLastPart - pszSrcDSName >= 3 && pszLastPart[-3] == ':' )
-                    pszLastPart -= 2;
-                CPLString osPrefix(pszSrcDSName);
-                osPrefix.resize(pszLastPart - pszSrcDSName);
-                pszSrcDSName = CPLStrdup((osPrefix +
-                    CPLProjectRelativeFilename(osVRTDirectory, pszLastPart)).c_str());
+                if( (osSrcDSName[nLastPart] == '/' || osSrcDSName[nLastPart] == '\\') &&
+                    nLastPart >= 3 && osSrcDSName[nLastPart-3] == ':' )
+                    nLastPart -= 2;
+                CPLString osPrefix(osSrcDSName);
+                osPrefix.resize(nLastPart);
+                osSrcDSName = osPrefix +
+                    CPLProjectRelativeFilename(osVRTDirectory, osSrcDSName.c_str() + nLastPart);
                 bDone = true;
                 break;
             }
         }
         if( !bDone )
         {
-            pszSrcDSName = CPLStrdup(
-                CPLProjectRelativeFilename(osVRTDirectory, pszSrcDSName));
+            osSrcDSName = CPLProjectRelativeFilename(osVRTDirectory, osSrcDSName.c_str());
         }
     }
-    else
-    {
-        pszSrcDSName = CPLStrdup(pszSrcDSName);
-    }
 
     // Are we accessing this datasource in shared mode?  We default
     // to shared for SrcSQL requests, but we also allow the XML to
@@ -618,7 +612,7 @@ bool OGRVRTLayer::FullInitialize()
     // Try to access the datasource.
 try_again:
     CPLErrorReset();
-    if( EQUAL(pszSrcDSName, "@dummy@") )
+    if( EQUAL(osSrcDSName.c_str(), "@dummy@") )
     {
         GDALDriver *poMemDriver =
             OGRSFDriverRegistrar::GetRegistrar()->GetDriverByName("Memory");
@@ -631,7 +625,7 @@ try_again:
     }
     else if( bSrcDSShared )
     {
-        if( poDS->IsInForbiddenNames(pszSrcDSName) )
+        if( poDS->IsInForbiddenNames(osSrcDSName.c_str()) )
         {
             CPLError(CE_Failure, CPLE_AppDefined,
                      "Cyclic VRT opening detected!");
@@ -645,7 +639,7 @@ try_again:
             if( bUpdate )
                 l_nFlags |= GDAL_OF_UPDATE;
             poSrcDS = (GDALDataset *)GDALOpenEx(
-                pszSrcDSName, l_nFlags, nullptr,
+                osSrcDSName.c_str(), l_nFlags, nullptr,
                 (const char *const *)papszOpenOptions, nullptr);
             CSLDestroy(papszOpenOptions);
             // Is it a VRT datasource?
@@ -666,7 +660,7 @@ try_again:
             if( bUpdate )
                 l_nFlags |= GDAL_OF_UPDATE;
             poSrcDS = (GDALDataset *)GDALOpenEx(
-                pszSrcDSName, l_nFlags, nullptr,
+                osSrcDSName.c_str(), l_nFlags, nullptr,
                 (const char *const *)papszOpenOptions, nullptr);
             CSLDestroy(papszOpenOptions);
             // Is it a VRT datasource?
@@ -701,13 +695,13 @@ try_again:
             CPLError(CE_Warning, CPLE_AppDefined,
                      "Cannot open datasource `%s' in update mode. "
                      "Trying again in read-only mode",
-                     pszSrcDSName);
+                     osSrcDSName.c_str());
             bUpdate = false;
             goto try_again;
         }
         if( strlen(CPLGetLastErrorMsg()) == 0 )
             CPLError(CE_Failure, CPLE_AppDefined,
-                     "Failed to open datasource `%s'.", pszSrcDSName);
+                     "Failed to open datasource `%s'.", osSrcDSName.c_str());
         goto error;
     }
 
@@ -745,14 +739,11 @@ try_again:
         {
             CPLError(CE_Failure, CPLE_AppDefined,
                      "Failed to find layer '%s' on datasource '%s'.",
-                     pszSrcLayerName, pszSrcDSName);
+                     pszSrcLayerName, osSrcDSName.c_str());
             goto error;
         }
     }
 
-    CPLFree(pszSrcDSName);
-    pszSrcDSName = nullptr;
-
     // Search for GeometryField definitions.
 
     // Create as many OGRVRTGeomFieldProps as there are
@@ -1120,7 +1111,6 @@ try_again:
 
 error:
     bError = true;
-    CPLFree(pszSrcDSName);
     poFeatureDefn->Release();
     poFeatureDefn = new OGRFeatureDefn(osName);
     poFeatureDefn->SetGeomType(wkbNone);
@@ -1161,24 +1151,23 @@ bool OGRVRTLayer::ResetSourceReading()
 
             const char *pszXField = poXField->GetNameRef();
             const char *pszYField = poYField->GetNameRef();
-            if( apoGeomFieldProps[i]->bUseSpatialSubquery )
+
+            OGRFieldType xType = poXField->GetType();
+            OGRFieldType yType = poYField->GetType();
+            if( !((xType == OFTReal || xType == OFTInteger ||
+                   xType == OFTInteger64) &&
+                  (yType == OFTReal || yType == OFTInteger ||
+                   yType == OFTInteger64)) )
             {
-                OGRFieldType xType = poXField->GetType();
-                OGRFieldType yType = poYField->GetType();
-                if( !((xType == OFTReal || xType == OFTInteger ||
-                       xType == OFTInteger64) &&
-                      (yType == OFTReal || yType == OFTInteger ||
-                       yType == OFTInteger64)) )
-                {
-                    CPLError(CE_Warning, CPLE_AppDefined,
-                             "The '%s' and/or '%s' fields of the source layer "
-                             "are not declared as numeric fields, "
-                             "so the spatial filter cannot be turned into an "
-                             "attribute filter on them",
-                             pszXField, pszYField);
-                    apoGeomFieldProps[i]->bUseSpatialSubquery = false;
-                }
+                CPLError(CE_Warning, CPLE_AppDefined,
+                         "The '%s' and/or '%s' fields of the source layer "
+                         "are not declared as numeric fields, "
+                         "so the spatial filter cannot be turned into an "
+                         "attribute filter on them",
+                         pszXField, pszYField);
+                apoGeomFieldProps[i]->bUseSpatialSubquery = false;
             }
+
             if( apoGeomFieldProps[i]->bUseSpatialSubquery )
             {
                 OGREnvelope sEnvelope;
@@ -1217,7 +1206,7 @@ bool OGRVRTLayer::ResetSourceReading()
 
                 if( !CPLIsInf(sEnvelope.MinX) )
                     osFilter +=
-                        CPLSPrintf("%s > %.15g", pszXField, sEnvelope.MinX);
+                        CPLSPrintf("\"%s\" > %.15g", pszXField, sEnvelope.MinX);
                 else if( sEnvelope.MinX > 0 )
                     osFilter += "0 = 1";
 
@@ -1226,7 +1215,7 @@ bool OGRVRTLayer::ResetSourceReading()
                     if( !osFilter.empty() )
                         osFilter += " AND ";
                     osFilter +=
-                        CPLSPrintf("%s < %.15g", pszXField, sEnvelope.MaxX);
+                        CPLSPrintf("\"%s\" < %.15g", pszXField, sEnvelope.MaxX);
                 }
                 else if( sEnvelope.MaxX < 0 )
                 {
@@ -1240,7 +1229,7 @@ bool OGRVRTLayer::ResetSourceReading()
                     if( !osFilter.empty() )
                         osFilter += " AND ";
                     osFilter +=
-                        CPLSPrintf("%s > %.15g", pszYField, sEnvelope.MinY);
+                        CPLSPrintf("\"%s\" > %.15g", pszYField, sEnvelope.MinY);
                 }
                 else if( sEnvelope.MinY > 0 )
                 {
@@ -1254,7 +1243,7 @@ bool OGRVRTLayer::ResetSourceReading()
                     if( !osFilter.empty() )
                         osFilter += " AND ";
                     osFilter +=
-                        CPLSPrintf("%s < %.15g", pszYField, sEnvelope.MaxY);
+                        CPLSPrintf("\"%s\" < %.15g", pszYField, sEnvelope.MaxY);
                 }
                 else if( sEnvelope.MaxY < 0 )
                 {
